######################################################################################################################
######################################################################################################################
######################################################################################################################
#
# iPF : Integrative Phenotyping Framework (iPF)
# Integrative Clustering of Multiple Omics Data Identifies Novel Lung Disease Subphenotypes
#
# Version : 1.1.0
# Authors : SungHwan Kim, Don Kang and George C. Tseng
# latest update : 09/09/2015
#
######################################################################################################################
######################################################################################################################
######################################################################################################################



iPF <- function(DList, distMat, Var.types, MDS=c("sammon"), MDS.tol= 1e-4, isParallel=FALSE, nCores=NULL, verbose=FALSE, reuseIPP=FALSE, method.cor=c("spearman"),
                .nx=NULL ,.nx2=rep(20,length(DList)), .nx2.too.far=rep(0.1,length(DList)), method.clustering=c("pam"), useCorDist=TRUE, is.Gap=FALSE,.std.axis=TRUE,.robust.coord=FALSE) {
  .p <- proto(expr = {
    .version <- "0.1.4"
    .verbose <- verbose
    .DTypes <- names(DList)
    .DList <- DList 
    .isParallel <- isParallel
    .workers <- NULL
    .dist <- NULL
    .mds <- NULL
    .mds.tol <- MDS.tol
    .coord.features <- list()
    #.gams <- vector(mode="list", length=length(DList))
    .reuseIPP <- reuseIPP
    .ippFull <- vector(mode="list", length=length(DList))
    .IPPdata <- NULL
    .PPPdata <- NULL
    .cluster <- NULL
    .nx <- NULL
    .nx2 <- .nx2
    .nx2.too.far <- .nx2.too.far
    .robust.coord <- .robust.coord
    #.gam.saved <- vector(mode="list", length=length(DList))
    .method.cor <- method.cor
    .useCorDist <- useCorDist
    #.colMeans <- list() #need to keep scale information for new data prediction
    #.colSds <- list()
    .is.Gap <- is.Gap
    .std.axis <- .std.axis
    Var.types <- Var.types
    requireAll(c('MASS','WGCNA','digest','cba','pcaMethods'))
    tryCatch(reassignInPackage2("image.matrix", pkgName="cba", NULL), error=function(e) invisible(NULL))
    
    if(isParallel) {
      if(!is.null(nCores))
        options(cores=nCores)
      if(.Platform$OS.type == "unix") {
        requireAll("doMC")
        registerDoMC()
        allowWGCNAThreads()
      } else { #windows
        requireAll("doSMP")
        if(is.null(getOption('cores')))
          options(cores=2)
        .workers <- startWorkers()
        registerDoSMP(.workers)
      }
    }
    
    Cleanup <- function(.) {
      if(!is.null(.$.workers))
        stopWorkers(.$.workers)
    }
    
    .initialize <- function(., .DList, .mds, .method.clustering, .method.cor) {
      stopifnot(all(length(names(.DList))>0) & all(!duplicated(names(.DList))))
      stopifnot(length(unlist(lapply(lapply(.DList, dimnames), function(x) sapply(x,length))))==2*length(.DList)) #no null dimnames
      stopifnot(length(table(sapply(.DList, nrow)))==1) #should have the same number of subjects
      stopifnot(all(sapply(.DList, nrow) - sapply(.DList, function(x) length(unique(rownames(x)))) == 0)) #full set of unique subjects
      stopifnot(all(apply(sapply(.DList, rownames), 1, function(x) length(unique(x)))==1)) #the same id among subjects
      stopifnot(!is.null(unlist(sapply(.DList, dimnames))))
      
      unlink(".gam", recursive=TRUE)
      .DList <- DList
      .$.DList0 <- .$.DList <- .DList
      .$.mds <- .mds
      .$.method.clustering <- .method.clustering
      .$.method.cor <- .method.cor
      
      for(i in 1:length(.DList))
        .$.DList[[i]] <- as.matrix(.DList[[i]]) #
      # .$.DList[[i]] <- scale(as.matrix(.DList[[i]])) #
      
      #                   for(i in 1:length(.DList)) {
      #                       .$.DList[[i]] <- as.matrix(.DList[[i]])
      #                       .$.colMeans[[i]] <- colMeans(.$.DList[[i]], na.rm=TRUE)
      #                       .$.colSds[[i]] <- colSds(.$.DList[[i]], na.rm=TRUE)
      #                       .$.DList[[i]] <- sweep(sweep(.$.DList[[i]], 2, .$.colMeans[[i]], "-"), 2, .$.colSds[[i]], "/") 
      #                   }
      names(.$.DList) <- names(.DList)
      #TODO... features existant in both data.. naming
      .dat <- do.call('cbind', .$.DList)
      
      .types <- NULL
      
      .$.dist <- distMat
      
      # .$.afterCalcDist()    ## <- note that I omit this step.
    }
    
    .ConvertIdx2RC <- function(., idx, mat) {
      col <- ceiling(idx / nrow(mat))
      row <- idx - nrow(mat)*(col-1)
      return(c(row,col))
    }
    
    .mixedDist <- function(., dat, Var.types, types, isAbs=FALSE) {
      
      #mat <- mixedDist( dat, Var.types, isAbs=FALSE)
      
      #as.dist(mat)
      
    }
    
    
    
    .calcIntegratedCoord <- function(., .dim=2) {
      stopifnot(.dim %in% 1:2)
      
      .coord <- do.call(.$.mds, list(.$.dist, k=.dim, tol=.$.mds.tol, trace=FALSE))$points
      
      .coord <- .coord[Var.types != "nom", ]
      
      .tmp <- 0
      for(i in 1:length(.$.DList)) {
        if(i>1)
          .tmp <- .tmp + ncol(.$.DList[[i-1]])
        .$.coord.features[[i]] <- .coord[.tmp + 1:ncol(.$.DList[[i]]),,drop=FALSE]
        
      }
      
      if(.$.std.axis){          
        for(idx in 1:length(.$.DList)){
          
          .$.coord.features[[idx]][,1] <- (as.numeric(.$.coord.features[[idx]][,1]) - min(as.numeric(.$.coord.features[[idx]][,1]))) / (max(as.numeric(.$.coord.features[[idx]][,1]))-min(as.numeric(.$.coord.features[[idx]][,1])))
          .$.coord.features[[idx]][,2] <- (as.numeric(.$.coord.features[[idx]][,2]) - min(as.numeric(.$.coord.features[[idx]][,2]))) / (max(as.numeric(.$.coord.features[[idx]][,2]))-min(as.numeric(.$.coord.features[[idx]][,2])))          
          
        }
      }  
      
      if(.$.robust.coord){
        for(idx in 1:length(.$.DList)){
          .dat <- .$.coord.features[[idx]]
          .dat[.$.coord.features[[idx]][ ,1] < quantile(.$.coord.features[[idx]][ ,1],0.005) | .$.coord.features[[idx]][ ,1] > quantile(.$.coord.features[[idx]][ ,1],0.995) ,1] <- NA
          .dat[.$.coord.features[[idx]][ ,2] < quantile(.$.coord.features[[idx]][ ,2],0.005) | .$.coord.features[[idx]][ ,2] > quantile(.$.coord.features[[idx]][ ,2],0.995) ,2] <- NA        
          .$.DList[[idx]] <- .$.DList[[idx]][,!(is.na(.dat[,1]) | is.na(.dat[,2]))]
          .$.coord.features[[idx]] <- na.omit(.dat)        
          
          
        }        
      }
    }
    
    
    .GamPred <- function(., x, n=100, xn=NULL, n2=20, xn2=NULL, too.far=NULL) {
      first <- x$smooth[[1]]$first.para
      last <- x$smooth[[1]]$last.para
      xs <- x$smooth[[1]]
      
      raw <- data.frame(x = as.numeric(x$model[xs$term[1]][[1]]), y = as.numeric(x$model[xs$term[2]][[1]])) #the same as .$.coord.features
      
      
      #n2 <- max(10, n2)
      if(is.null(xn2)) {
        xm <- seq(min(raw$x), max(raw$x), length = n2)
        ym <- seq(min(raw$y), max(raw$y), length = n2)
      } else {          
        
        stopifnot(length(xn2)==2)
        xm <- xn2$x
        ym <- xn2$y
        
        # xm and ym are the same as xn2's component in the list. In other words it is from MDS coordinates
        # xn2 <- list(x=seq(from=min(.$.coord.features[[idx]][,1]), to=max(.$.coord.features[[idx]][,1]), length.out=.nx2[idx]),
        #             y=seq(from=min(.$.coord.features[[idx]][,2]), to=max(.$.coord.features[[idx]][,2]), length.out=.nx2[idx]))
      }
      
      # xm and ym are the same as xn2's component in the list.
      # xn2 <- list(x=seq(from=min(.$.coord.features[[idx]][,1]), to=max(.$.coord.features[[idx]][,1]), length.out=.nx2[idx]),
      #             y=seq(from=min(.$.coord.features[[idx]][,2]), to=max(.$.coord.features[[idx]][,2]), length.out=.nx2[idx]))
      ## n2 is the same as nx2 (=40, as default)
      xx <- rep(xm, n2)
      yy <- rep(ym, rep(n2, n2))
      
      if (too.far > 0){            
        exclude <- mgcv:::exclude.too.far(xx, yy, raw$x, raw$y, dist = too.far)
        ## raw$x is the divided MDS coordinate into 20 groups.          
      }else{
        exclude <- rep(FALSE, n2 * n2)
      }
      
      .dat <- data.frame(x = xx, y = yy)
      names(.dat) <- c(xs$term[1], xs$term[2])
      X <- PredictMat(xs, .dat)
      p <- x$coefficients[first:last]
      fit <- X %*% p
      fit[exclude] <- NA
      
      
      return(fit)
    }
    
    .PredictMat <- function(., object,data,n=nrow(data),bypass=FALSE)
      ## wrapper function which calls Predict.matrix and imposes same constraints as 
      ## smoothCon on resulting Prediction Matrix
    { X <- Predict.matrix2(object,data)
      if (is.null(attr(X,"by.done"))) { ## handle `by variables' 
        if (object$by!="NA")  # deal with "by" variable 
        { by <- get.var(object$by,data)
          if (is.null(by)) stop("Can't find by variable")
          if (is.factor(by)) {
            by.dum <- as.numeric(object$by.level==by)
            X <- by.dum*X
          } else { 
            if (length(by)!=nrow(X)) stop("`by' variable must be same dimension as smooth arguments")
            X <- as.numeric(by)*X
          }
        }
      }
      attr(X,"by.done") <- NULL
      offset <- attr(X,"offset")
      
      if(!bypass) {
        ## now deal with any necessary model matrix summation
        if (n != nrow(X)) {
          q <- nrow(X)/n ## note: can't get here if `by' a factor
          ind <- 1:n 
          Xs <- X[ind,]
          for (i in 2:q) {
            ind <- ind + n
            Xs <- Xs + X[ind,]
          }
          X <- Xs
        }
      }
      
      qrc <- attr(object,"qrc")
      if (!is.null(qrc)) { ## then smoothCon absorbed constraints
        j <- attr(object,"nCons")
        if (j>0) { ## there were constraints to absorb - need to untransform
          k<-ncol(X)
          if (inherits(qrc,"qr")) {
            indi <- attr(object,"indi") ## index of constrained parameters
            if (is.null(indi)) {
              if (sum(is.na(X))) {
                ind <- !is.na(rowSums(X))
                X1 <- t(qr.qty(qrc,t(X[ind,,drop=FALSE]))[(j+1):k,,drop=FALSE]) ## XZ
                X <- matrix(NA,nrow(X),ncol(X1))
                X[ind,] <- X1
              } else {
                X <- t(qr.qty(qrc,t(X))[(j+1):k,,drop=FALSE])
              }
            } else { ## only some parameters are subject to constraint
              nx <- length(indi)
              nc <- j;nz <- nx - nc
              if (sum(is.na(X))) {
                ind <- !is.na(rowSums(X))
                X[ind,indi[1:nz]]<-t(qr.qty(qrc,t(X[ind,indi,drop=FALSE]))[(nc+1):nx,])
                X <- X[,-indi[(nz+1):nx]]
                X[!ind,] <- NA 
              } else { 
                X[,indi[1:nz]]<-t(qr.qty(qrc,t(X[,indi,drop=FALSE]))[(nc+1):nx,,drop=FALSE])
                X <- X[,-indi[(nz+1):nx]]
              }
            }
          } else if (qrc>0) { ## simple set to zero constraint
            X <- X[,-qrc]
          } else if (qrc<0) { ## params sum to zero
            X <- t(diff(t(X)))
          }
        }
      }
      ## drop columns eliminated by side-conditions...
      del.index <- attr(object,"del.index") 
      if (!is.null(del.index)) X <- X[,-del.index]
      attr(X,"offset") <- offset
      X
    } ## end of PredictMat
    
    
    doPhenotyping <- function(., k, ik,.dim=2, redoMDS=FALSE, 
                              doCV=FALSE, .nxs=NULL, .nx2s=NULL, .nx2s.too.far=NULL, pct=.5, B=100, seed=NULL, ...) {
      stopifnot(.dim %in% 1:2)
      
      if(length(.$.coord.features) < length(.$.DList) | redoMDS | any(sapply(.$.coord.features,ncol)!=.dim))  ##  This should be certainly true for calculating coordinate information
        .$.calcIntegratedCoord(.dim)
      
      #if(doCV) {
      #  .res <- .$.doCV(k, .dim, .nxs=.nxs, .nx2s=.nx2s, .nx2s.too.far=.nx2s.too.far, pct=pct, B=B, seed=seed)
      
      #  if(.dim==1) {
      #    .$.nx <- .res$.nx
      #  } else {
      #    .$.nx2 <- .res$.nx2
      #    .$.nx2.too.far <- .res$.nx2.too.far
      #  }
      #}
      
      .$.IPPdata <- .$.doIPP(.dim)
      
      names(.$.IPPdata) <- names(.$.DList)
      .$.findPatterns(k=k, ...)
      #  .$.integative.findPatterns(ik=ik, ...)
    }
    
    
    
    
    .doIPP <- function(., .dim=2, .DList=NULL, .name.data=NULL, .nx=NULL, .xn=NULL, .nx2=NULL, .xn2=NULL, .nx2.too.far=NULL) { #Individual Phenomic Presentation
      stopifnot(.dim %in% 1:2)
      if(is.null(.$.coord.features))
        .$.calcIntegratedCoord(.dim)
      if(is.null(.DList)){
        .DList <- .$.DList
      } else {
        for(i in 1:length(.DList))
          .DList[[i]] <- scale(as.matrix(.DList[[i]]))
      }
      
      if(is.null(.nx2)) .nx2 <- .$.nx2 else if(length(.nx2)==1) .nx2 <- rep(.nx2, length(.DList))
      if(is.null(.nx2.too.far)) .nx2.too.far <- .$.nx2.too.far else if(length(.nx2.too.far)==1) .nx2.too.far <- rep(.nx2.too.far, length(.DList))
      if(is.null(.name.data)) {
        .idx.data <- 1:length(.DList)
      } else {
        stopifnot(all(.name.data %in% names(.DList)))
        .idx.data <- match(.name.data, names(.DList))
      }
      
      foreach(idx=.idx.data) %do% {
        
        ..xn2 <- list(x=seq(from=min(.$.coord.features[[idx]][,1]), to=max(.$.coord.features[[idx]][,1]), length.out=.nx2[idx]),
                      y=seq(from=min(.$.coord.features[[idx]][,2]), to=max(.$.coord.features[[idx]][,2]), length.out=.nx2[idx]))
        
        
        
        .ipp <- t(foreach(j=1:nrow(.DList[[idx]]), .combine=cbind) %dopar% {
          .dat <- cbind.data.frame(.DList[[idx]][j,], .$.coord.features[[idx]]) 
          
          
          
          colnames(.dat) <- c('y', paste('x',1:ncol(.$.coord.features[[idx]]),sep=''))
          
          .gam <- .getIndivGam(ridx=j, .dat=.dat, .name.data=names(.DList)[idx], .dim=.dim)
          .$.GamPred(.gam, n2=.nx2[idx], xn2=..xn2, too.far=.nx2.too.far[idx])          
        })
        
        rownames(.ipp) <- rownames(.DList[[idx]])
        return(.ipp)
      }
    }
    
    
    
    
    .predict <- function(., ClusterInfo, .dim, .DList) {
      
      if(.$.robust.coord){
        for(idx in 1:length(.$.DList)){
          .DList[[idx]] <- .DList[[idx]][ ,colnames(.DList[[idx]])  %in% colnames(.$.DList[[idx]])]
        }        
      }
      
      stopifnot(all(names(.DList) %in% names(.$.DList)))
      .idx.data <- match(names(.DList), names(.$.DList))
      
      .ipp <- .$.doIPP(.dim, .DList)
      
      .pred <- foreach(i=1:length(.idx.data), .combine=cbind) %dopar% {
        .isRow <- names(.$.DList)[i]==names(dimnames(ClusterInfo$table))[1]
        if(.isRow)
          .mems <- sapply(ClusterInfo$member, unlist)   ## This works for collapsing samples with respect to the first dataset, say, 'clinical'.
        else
          .mems <- lapply(1:ncol(ClusterInfo$table), function(i) unlist(lapply(ClusterInfo$member, function(x) x[i])))
        ## This works for collapsing samples with respect to the second dataset, say, 'gene expression'.
        
        .cpp <- .$.getCPP2(.DList=.$.DList, .dim=.dim, .mems=.mems, .idx.data=i, thoroughly=FALSE)
        #########################################
        ## note that .$.DList is training dataset
        #########################################
        
        apply(dist(.ipp[[i]], t(do.call('cbind', unlist(.cpp,recursive=F)))), 1, which.min)
      }
      
      colnames(.pred) <- names(.DList)
      list(ipp=.ipp, pred=.pred)
    }
    
    #with(lgrc, debug(.findPatterns)); lgrc$.findPatterns(k=3)
    .findPatterns <- function(., k, ...) {
      if(is.null(.$.IPPdata)){
        .$.IPPdata <- .$.doIPP(.dim=ncol(.$.coord.features[[1]]))
        names(.$.IPPdata) <- names(.$.DList)                     
      }
      
      if(.$.is.Gap){
        .pam1 <- function(x,k) list(cluster = pam(x,k, cluster.only=TRUE))                  
        #gsPam1 <- clusGap(as.matrix(t(do.call(cbind, .$.IPPdata))), FUN = .pam1, K.max = 5, B = 500)            
        gsPam.single <- foreach(i = 1:length(.$.IPPdata)) %do% {
          ..cor <- cor(t(.$.IPPdata[[i]]),use="pairwise.complete.obs") 
          gsPam1 <- clusGap(as.matrix(as.dist(1-..cor)), FUN = .pam1, K.max = 10, B = 100)                
          .ik <- which(gsPam1$Tab[,3] == max(gsPam1$Tab[,3]))
          return(list(gsPam1=gsPam1 ,.ik=.ik))
        }
        .$gsPam.single <- gsPam.single
        .k <- do.call(c,lapply(unlist(.$gsPam.single,recursive=F), function(x) if(length(x)==1) x ))
        # .k <- lapply(.$gsPam.single, function(x)
      } else {.k <- k 
              .$.k <- k
      }
      
      
      .$.cluster <- foreach(i=1:length(.$.IPPdata)) %dopar% {
        if(.$.method.clustering=="pam") {
          if(.$.useCorDist)
            .res <- do.call(.$.method.clustering, list(x=as.dist(1-cor(t(.$.IPPdata[[i]]), use="pairwise.complete.obs")), k=.k[i], ...))
          else
            .res <- do.call(.$.method.clustering, list(x=.$.IPPdata[[i]], k=.k[i], ...))
          
          .res$centers <- foreach(j=1:length(table(.res$clustering)), .combine=rbind) %do% {
            colMedians(.$.IPPdata[[i]][which(.res$clustering==j),], na.rm=TRUE)
          }
          #.res$centers <- .$.IPPdata[[i]][.res$id.med,]
          return(.res)
        } 
      }
      names(.$.cluster) <- names(.$.IPPdata)      
    }
    
    .integative.findPatterns <- function(., ik, ...) {
      if(is.null(.$.IPPdata)){
        .$.IPPdata <- .$.doIPP(.dim=ncol(.$.coord.features[[1]]))
        names(.$.IPPdata) <- names(.$.DList)                     
      }
      
      if(.$.is.Gap){
        .pam1 <- function(x,k) list(cluster = pam(x,k, cluster.only=TRUE))                  
        #gsPam1 <- clusGap(as.matrix(t(do.call(cbind, .$.IPPdata))), FUN = .pam1, K.max = 5, B = 500)      
        
        ..cor <- matrix(apply(do.call(rbind, lapply(lapply(.$.IPPdata, function(x) cor(t(x),,use="pairwise.complete.obs")),as.vector)),2,mean) , dim(.$.IPPdata[[1]])[1],dim(.$.IPPdata[[1]])[1])
        gsPam1 <- clusGap(as.matrix(as.dist(1-..cor)), FUN = .pam1, K.max = 10, B = 100)
        
        .ik <- which(gsPam1$Tab[,3] == max(gsPam1$Tab[,3]))
        .$gsPam <- gsPam1
      }
      else {.ik <- ik}
      
      ..cor <- matrix(apply(do.call(rbind, lapply(lapply(.$.IPPdata, function(x) cor(t(x),,use="pairwise.complete.obs")),as.vector)),2,mean) , dim(.$.IPPdata[[1]])[1],dim(.$.IPPdata[[1]])[1])
      .$integrated.cluster <- do.call(.$.method.clustering, list(x=as.dist(1-..cor), k= .ik, ...))      
      
      #.$integrated.cluster <- do.call(.$.method.clustering, list(x=as.dist(1-cor(t(do.call(cbind, .$.IPPdata)),use="pairwise.complete.obs")), k= .ik, ...))
      #    .$integrated.cluster$centers <- foreach(j=1:length(table(.$integrated.cluster$clustering)), .combine=rbind) %do% {
      #      colMedians(cbind(.$.IPPdata[[1]], .$.IPPdata[[2]])[which(.$integrated.cluster$clustering==j),], na.rm=TRUE)
      #    }    
    }
    
    
    adjustClusterInfo <- function(., ClusterInfo, ord=NULL, x.ord=NULL, y.ord=NULL) {
      if(!is.null(ord)) 
        x.ord <- y.ord <- ord
      if(!is.null(x.ord)) {
        stopifnot(length(x.ord)==ncol(ClusterInfo$table) & all(x.ord %in% 1:ncol(ClusterInfo$table)) & all(1:ncol(ClusterInfo$table) %in% x.ord))
        .ord <- match(x.ord, colnames(ClusterInfo$table)) 
        ClusterInfo$table <- ClusterInfo$table[,.ord]
        ClusterInfo$member <- sapply(1:length(ClusterInfo$member), function(i) lapply(ClusterInfo$member[i], function(x) x[.ord]))
        ClusterInfo$what <- sapply(1:length(ClusterInfo$what), function(i) lapply(ClusterInfo$what[i], function(x) x[.ord]))
      }
      if(!is.null(y.ord)) {
        stopifnot(length(y.ord)==nrow(ClusterInfo$table) & all(y.ord %in% 1:nrow(ClusterInfo$table)) & all(1:nrow(ClusterInfo$table) %in% y.ord))
        .ord <- match(y.ord, rownames(ClusterInfo$table))
        ClusterInfo$table <- ClusterInfo$table[.ord,]
        ClusterInfo$member <- ClusterInfo$member[.ord] 
        ClusterInfo$what <- ClusterInfo$what[.ord] 
      }
      return(ClusterInfo)
    } 
    
    #when more than 2 data sets
    #orderByWhat=c(NULL,"Desc","Incr")
    
    
    
    #diagnosis: string vector/ names are subjects' id's
    #.which: index of DList
    #classes: additional subjects grouping such as diagnosis
    #TODO.. apply substitue for classes variables
    getClusterInfo <- function(., .DList=NULL, .name.data=NULL, what=NULL, .flip=FALSE, .flip.x=FALSE, .flip.y=FALSE) {
      if(is.null(.$.cluster))
        .$doPhenotyping()
      if(is.null(.DList)) .DList <- .$.DList
      
      if(is.null(.name.data)) {
        stopifnot(length(.DList)<=2)
        .name.data <- names(.DList)
      }
      
      stopifnot(length(.name.data) %in% 1:2)
      stopifnot(all(.name.data %in% names(.DList)))
      
      if(!is.null(what)) {
        what <- what[match(rownames(.DList[[1]]), names(what), nomatch=0)]
        stopifnot(length(what)==nrow(.DList[[1]]))
        stopifnot(all(rownames(.DList[[1]]) %in% unique(names(what))))
      }
      
      .idx.data <- match(.name.data, names(.DList))
      
      ### match cluster number among data sets by their similarity
      #order for each data set
      #                   .clusterOrder <- foreach(idx=.idx.data) %do% {
      #                       order(sammon(dist(.$.cluster[[idx]]$centers), k=1, trace=FALSE)$points) #only consider physical distance rather than pattern                        
      #                   }
      
      .dim <- ncol(.$.coord.features[[1]])
      
      
      .nx2 <- max(.$.nx2)
      .common.x <- apply(sapply(.$.coord.features, function(x) range(x[,1])), 1, function(x) x[which.max(abs(x))])
      .common.y <- apply(sapply(.$.coord.features, function(x) range(x[,2])), 1, function(x) x[which.max(abs(x))])
      .xn2 <- list(x=seq(from=.common.x[1], to=.common.x[2], length.out=.nx2), y=seq(from=.common.y[1], to=.common.y[2], length.out=.nx2))
      
      .dat <- foreach(idx=.idx.data, .combine=cbind) %do% {
        #.$.getCPP2(.DList=.DList, .dim=.dim, .mems=lapply(split(.$.cluster[[idx]]$cluster,.$.cluster[[idx]]$cluster),names), xn2=.xn2,
        #           .idx.data=idx, thoroughly=FALSE, too.far=.$.nx2.too.far[idx]) 
        do.call(cbind, sapply(.$.getCPP2(.DList=.DList, .dim=2 , .mems=lapply(split(.$.cluster[[idx]]$cluster,.$.cluster[[idx]]$cluster),names), .idx.data=idx 
                                         ,thoroughly=FALSE, too.far=.$.nx2.too.far[idx]),c) )
      }
      
      
      
      .loc <- sammon(dist(t(.dat)), k=1, trace=FALSE)$points
      k <- length(table(.$.cluster[[1]]$cluster)) #TODO.. in general two number of clusters may be different
      
      
      .clusterOrder <- list(1:.$.k[1], 1:.$.k[2] )
      
      #.clusterOrder <- foreach(i=1:length(.idx.data)) %do% {
      #  order(.loc[k*(i-1) + 1:k])
      #}
      
      if(.flip) {
        .flip.x <- .flip.y <- TRUE
      }
      
      if(.flip.y)
        .clusterOrder[[1]] <- rev(.clusterOrder[[1]])
      if(.flip.x)
        .clusterOrder[[2]] <- rev(.clusterOrder[[2]])
      
      if(length(.idx.data)==1) {
        .clsRes <- table(.$.cluster[[.idx.data]]$cluster)
        .clsRes <- .clsRes[.clusterOrder[[.idx.data]]]
        .member <- foreach(i=1:nrow(.clsRes)) %do% {
          rownames(.DList[[1]])[which(.$.cluster[[.idx.data]]$cluster==.clusterOrder[[.idx.data]][i])]
        }
        names(dimnames(.clsRes)) <- names(.DList)[.idx.data]
        .res <- list(table=.clsRes, member=.member)
        if(!is.null(what)) {
          .what <- foreach(i=1:nrow(.clsRes)) %do% {
            what[match(rownames(.DList[[1]])[which(.$.cluster[[.idx.data[1]]]$cluster==.clusterOrder[[.idx.data[1]]][i])], 
                       names(what))]
          }
          .res$what=.what
        }
      } else {
        .clsRes <- table(.$.cluster[[.idx.data[1]]]$cluster, .$.cluster[[.idx.data[2]]]$cluster)
        .clsRes <- .clsRes[.clusterOrder[[.idx.data[1]]], .clusterOrder[[.idx.data[2]]]]
        .member <- foreach(i=1:nrow(.clsRes)) %do% {
          foreach(j=1:ncol(.clsRes)) %do% {
            rownames(.DList[[1]])[which(.$.cluster[[.idx.data[1]]]$cluster==.clusterOrder[[.idx.data[1]]][i] &
                                          .$.cluster[[.idx.data[2]]]$cluster==.clusterOrder[[.idx.data[2]]][j])]
          }
        }
        names(dimnames(.clsRes)) <- names(.DList)[.idx.data]
        .res <- list(table=.clsRes, member=.member)
        if(!is.null(what)) {
          .what <- foreach(i=1:nrow(.clsRes)) %do% {
            foreach(j=1:ncol(.clsRes)) %do% {
              what[match(rownames(.DList[[1]])[which(.$.cluster[[.idx.data[1]]]$cluster==.clusterOrder[[.idx.data[1]]][i] & 
                                                       .$.cluster[[.idx.data[2]]]$cluster==.clusterOrder[[.idx.data[2]]][j])], 
                         names(what))]
            }
          }
          .res$what=.what
        }
      }
      
      return(.res)
    } 
    
    #TODO.. combine box and pie
    
    
    #ClusterInfo object can be modified, e.g. adding additional information using member information (refer to classes) 
    pieChart <- function(., ClusterInfo, what="diagnosis", what.order=NULL, .legend=FALSE, .color=NULL, 
                         which.row=NULL, which.col=NULL, collapse.row=FALSE, collapse.col=FALSE, .flip=FALSE,
                         .save=FALSE, .image.format="png", .file.name=paste("Rplots",.image.format,sep="."), .width=800, .height=800, .res=300, .compress=FALSE) {
      stopifnot(!is.null(ClusterInfo$what))
      stopifnot(length(dim(ClusterInfo$table))<=2)
      
      if(collapse.row) {
        .tab <- matrix(apply(ClusterInfo$table,2,sum),nrow=1)
        ClusterInfo$table <- .tab
        ClusterInfo$what <- lapply(1:ncol(ClusterInfo$table), function(i) unlist(lapply(ClusterInfo$what, function(x) x[i])))
      }
      if(collapse.col) {
        .tab <- matrix(apply(ClusterInfo$table,1,sum),ncol=1)
        ClusterInfo$table <- .tab
        ClusterInfo$what <- lapply(ClusterInfo$what, unlist)
      }
      
      .nrow <- dim(ClusterInfo$table)[1]
      .ncol <- dim(ClusterInfo$table)[2]
      
      .idx <- 1:length(ClusterInfo$table)
      if(!collapse.row & !is.null(which.row)) {
        .idx <- (.ncol*(which.row-1)+1):(.ncol*which.row)
      }
      if(!collapse.col & !is.null(which.col)) {
        .idx.c <- seq(which.col, by=.ncol, length.out=.nrow)
        .idx <- intersect(.idx, .idx.c)
      }
      
      if(.flip) {
        .tmp <- .nrow; .nrow <- .ncol; .ncol <- .tmp
      }
      
      pieData <- data.frame(group=rep(.idx, times=t(ClusterInfo$table)[.idx]), 
                            what=unlist(sapply(ClusterInfo$what, c)[.idx]))
      
      colnames(pieData)[2] <- what
      
      if(is.null(what.order))
        pieData[[what]] <- factor(pieData[[what]])
      else {
        stopifnot(all(unique(pieData[[what]]) %in% what.order))
        pieData[[what]] <- factor(pieData[[what]], levels=what.order)
      }
      
      pieData <- table(pieData$group, pieData[[what]]) 
      
      .empty <- setdiff(1:(.nrow * .ncol), rownames(pieData))
      if(length(.empty)>0) {
        for(emp in .empty) {
          if(emp==1) {
            pieData <- rbind(rep(0,ncol(pieData)), pieData)
          } else {
            pieData <- rbind(pieData[1:(emp-1),], rep(0,ncol(pieData)), pieData[emp:nrow(pieData),])
          }
        }
        rownames(pieData) <- 1:(.nrow * .ncol)   
      }
      
      if(is.null(.color)) {
        if(ncol(pieData)<3) {
          .color <- RColorBrewer::brewer.pal(3, 'Spectral')
          if(ncol(pieData)==2)
            .color <- .color[c(1,3)]
          else
            .color <- .color[1]
        } else {
          .color <- RColorBrewer::brewer.pal(ncol(pieData), 'Spectral')
        }
      }
      
      if(.save & .image.format != "pdf")
        do.call(.image.format, list(filename=.file.name, width=.width*.ncol, height=.height*.nrow, res=.res))
      
      #FIXME.. how to handle empty pie chart!
      pie <- ggplot(as.data.frame.table(pieData), aes(x=factor(1), y=Freq, fill=Var2)) + geom_bar(width = 1) + 
        scale_fill_manual(values=.color) +
        facet_wrap(~Var1, nrow=.nrow, ncol=.ncol, scales = "free_y", drop=FALSE) + coord_polar(theta = "y") 
      
      pie <- pie + opts(axis.ticks=theme_blank(), axis.title.x=theme_blank(), axis.title.y=theme_blank(), 
                        axis.text.y=theme_blank(), axis.text.x=theme_text(face="bold", size=20), 
                        plot.margin = unit(c(-.5,-.5,-.5,-.5), "lines"), panel.background = theme_blank(), 
                        strip.background=theme_blank(), strip.text.x=theme_blank())
      if(!.legend)
        pie <- pie + opts(legend.position="none")
      pie$coordinates$guide_background <- guide_background
      pie$coordinates$guide_foreground <- guide_foreground3
      print(pie)
      
      if(.save) {
        dev.off()
        if(.compress)
          R.utils::gzip.default(.file.name, overwrite=T, remove=T)
      }
    }
    
    
    
    .pseudo.colors <- c(rgb(0,0,seq(.5625,1,.0625)),
                        rgb(0,seq(.0625,1,.0625),1),
                        rgb(seq(.0625,1,.0625), 1, seq(.9375,0,-.0625)),
                        rgb(1,seq(.9375,0,-.0625),0),
                        rgb(seq(.9375,.5,-.0625),0,0))
    
    
    
    
    .getCPP2 <- function(. ,.DList, .dim, .mems, .idx.data, thoroughly, 
                         xn=NULL, too.far=NULL, pointPred=FALSE) {
      foreach(clsID=1:length(.mems)) %dopar% { #cluster level
        m <- .mems[[clsID]]
        foreach(idx=.idx.data) %do% { #dataset level
          
          .nx2 <- .$.nx2[idx]
          .xn2 <- list(x=seq(from=min(.$.coord.features[[idx]][,1]), to=max(.$.coord.features[[idx]][,1]), length.out=.nx2),
                       y=seq(from=min(.$.coord.features[[idx]][,2]), to=max(.$.coord.features[[idx]][,2]), length.out=.nx2))
          
          .name <- "Cluster"
          apply(.$.IPPdata[[idx]][match(m,rownames(.DList[[idx]])),],2, function(x) mean(x,na.rm=T) )  
        }
      }
    }
    
    
    
    
    
    
    #TODO.. change thoroughly to reuseIPP
    Plot <- function(., ClusterInfo=NULL, .dim=1, .DList=NULL, id=NULL, .name.data=NULL, thoroughly=ifelse(is.null(id),FALSE,TRUE),
                     .colors.1d.cluster=NULL, .colors.1d.data=NULL, .ylim.1d.range=c(-2,2), .color.2d.range=c(-2,2), .lwd=3, .lwd.indiv=.1, #.cex.points=.05, .alpha.points=1/10,
                     which.row=NULL, which.col=NULL, collapse.row=FALSE, collapse.col=FALSE, .flip=FALSE, .nopar=FALSE, .mems=NULL, .ncol=NULL, .nrow=NULL, 
                     .save=FALSE, .image.format="png", .file.name=paste("Rplots",.image.format,sep="."), .width=400, .height=400, .res=300, .compress=FALSE) {
      stopifnot(.dim %in% 1:2)
      
      if(is.null(.DList)){
        .DList <- .$.DList
      } 
      else {
        for(i in 1:length(.DList))
          .DList[[i]] <- scale(as.matrix(.DList[[i]]))
      }
      
      if(is.null(.name.data))
        .name.data <- names(.DList)
      else
        stopifnot(all(.name.data %in% names(.DList)))
      if(.dim==2) stopifnot(!is.null(.name.data) & length(.name.data)==1)
      .idx.data <- match(.name.data, names(.DList))
      
      if(is.null(.colors.1d.data)) {
        if(is.null(.name.data))
          .colors.1d.data <- 1:length(.DList) + 1 #no black
        else
          .colors.1d.data <- 1:length(.name.data) + 1
      } else {
        if(is.null(.name.data)) 
          stopifnot(length(.colors.1d.data)==length(.DList))
        else stopifnot(length(.colors.1d.data)==length(.name.data))
      }
      
      if(is.null(id)) { #cluster plot
        if(is.null(.mems)) {
          if(collapse.row) {
            .tab <- matrix(apply(ClusterInfo$table,2,sum),nrow=1)
            ClusterInfo$table <- .tab
            ClusterInfo$member <- lapply(1:ncol(ClusterInfo$table), function(i) unlist(lapply(ClusterInfo$member, function(x) x[i])))
          }
          if(collapse.col) {
            .tab <- matrix(apply(ClusterInfo$table,1,sum),ncol=1)
            ClusterInfo$table <- .tab
            ClusterInfo$member <- lapply(ClusterInfo$member, unlist)
          }
          
          if(collapse.row | collapse.col)
            .mems <- ClusterInfo$member
          else
            .mems <- unlist(ClusterInfo$member,recursive=F)
        }
        
        if(is.null(.nrow))
          .nrow <- dim(ClusterInfo$table)[1]
        if(is.null(.ncol))
          .ncol <- dim(ClusterInfo$table)[2]
        
        # .ys <- .$.getCPP(.DList=.DList, .dim=.dim, .mems=.mems, .idx.data=.idx.data, thoroughly=thoroughly)
        
        .ys <-  .$.getCPP2(.DList=.DList, .dim=.dim, .mems=.mems, .idx.data=.idx.data, thoroughly=thoroughly)
        
        
        
      } else { #individual plot
        stopifnot(id %in% rownames(.DList[[1]]))
        .mems <- id
        .nrow <- .ncol <- 1
      }
      
      if(.flip) {
        .tmp <- .nrow; .nrow <- .ncol; .ncol <- .tmp
      }
      
      
      if(.save & .image.format != "pdf") 
        do.call(.image.format, list(filename=.file.name, width=.width*.ncol, height=.height*.nrow, res=.res))
      
      if(!.nopar)
        par(mfrow=c(1,1), cex=1, mar=c(0,0,0,0),bg="grey79") # bottom,left,top,right
      
      
      gridX <- seq(min(.$.coord.features[[.idx.data]][,1]), max(.$.coord.features[[.idx.data]][,1]), length=.$.nx2[.idx.data])
      gridY <- seq(min(.$.coord.features[[.idx.data]][,2]), max(.$.coord.features[[.idx.data]][,2]), length=.$.nx2[.idx.data])
      #gridX <- seq(0,1, length=.$.nx2[.idx.data])
      #gridY <- seq(0,1, length=.$.nx2[.idx.data])
      
      if(is.null(id)) {
        for(j in 1:length(.mems)) { #cluster level
          png(filename=paste("Rplots_new",j, ".png",sep="") ,width=830, height=1000, res=300)
          
          par(bg = "gray89")
          plot.new()
          #par(mfrow=c(1,1), cex=1, mar=c(0,0,0,0))
          .filled.contour(gridX, gridY, matrix(.ys[[j]][[1]],.$.nx2[.idx.data],.$.nx2[.idx.data]), 
                          levels=c(min(matrix(.ys[[j]][[1]],.$.nx2[.idx.data],.$.nx2[.idx.data]),.color.2d.range[1]-1,na.rm=T), 
                                   seq(.color.2d.range[1], .color.2d.range[2], length=length(.$.pseudo.colors)), 
                                   max(matrix(.ys[[j]][[1]],.$.nx2[.idx.data],.$.nx2[.idx.data]),.color.2d.range[2]+1,na.rm=T)), 
                          col=c(.$.pseudo.colors[1],.$.pseudo.colors,.$.pseudo.colors[64]))
          dev.off(); R.utils::gzip.default(paste("Rplots_new",j, ".png",sep=""), overwrite=T, remove=T)
        }
      } 
      
    
      if(.save) {
        dev.off()
        if(.compress)
          R.utils::gzip.default(.file.name, overwrite=T, remove=T)
      }
    }
    
    
    
     
  }
  )
  
  MDS <- match.arg(MDS)
  method.clustering <- match.arg(method.clustering)
  method.cor <- match.arg(method.cor)
  
  .p$.initialize(.DList=DList, .mds=MDS, .method.clustering=method.clustering, .method.cor=method.cor)
  
  return(.p)
}

.getIndivGam <- function(ridx, .dat, .name.data, .dim) {
  
  .gam <- gam(as.formula(paste("y ~ ", "s(", paste(colnames(.dat)[-1], collapse=","), ",bs='tp')", sep="")), data=.dat)
  
  #       if(.dim==1)
  #           .gam <- gam(y ~ s(x1,bs="tp"), data=.dat) 
  #       else if(.dim==2)
  #           .gam <- gam(y ~ s(x1,x2,bs="tp"), data=.dat)
  # if(!file.exists(".gam")) 
  #    dir.create(".gam")
  #  save(.gam, file=.filename)
  
  return(.gam)
}


TissueID2NewID <- function(tid, na.rm=TRUE) {
  #   require(RMySQL)
  #   con <- dbConnect(MySQL(), user="root", dbname="MP7second")
  #   res <- dbGetQuery(con, "SELECT newid FROM DIAGNOSE where tisslbl in (842, 1098)")
  
  if(!exists("DIAGNOSE"))
    load(file="DIAGNOSE.rda")
  mode(tid) <- "integer"
  tid[is.na(tid)] <- 0
  nid <- DIAGNOSE$newid[match(tid, DIAGNOSE$tisslbl)]
  if(na.rm)
    return(na.omit(nid))
  else
    return(nid)
}


requireAll <- function(packages) {
  .packages <- setdiff(packages, installed.packages()[,'Package'])
  if(length(.packages)>0) {
    source("http://bioconductor.org/biocLite.R")
    biocLite(.packages, dependencies=TRUE)
  }
  for(package in packages)
    do.call(require, list(package))
}



piechart.new <- function(clsInfo){
  ClusterInfo<- clsInfo
  .nrow <- dim(ClusterInfo$table)[1]
  .ncol <- dim(ClusterInfo$table)[2]
  .idx <- 1:length(ClusterInfo$table)
  pieData <- data.frame(group=rep(.idx, times=t(ClusterInfo$table)[.idx]), 
                        what=unlist(sapply(ClusterInfo$what, c)[.idx]))
  
  what="diagnosis"
  colnames(pieData)[2] <- what
  
  diag.order <- c("COPD", "Respiratory bronchiolitis", "DIP", "Fibrosis-Uncharacterized", "Hypersensitivity Pneumonitis", "NSIP", "Other ILD", "Idiopathic UIP")
  what.order=diag.order
  pieData[[what]] <- factor(pieData[[what]], levels=what.order)
  pieData <- table(pieData$group, pieData[[what]])
  
  piedata <- as.data.frame.table(pieData)
  
  .tmp <-rep(0,nrow(piedata))
  for(i in 1:max(as.numeric(as.matrix(piedata[,1])))){
    tmp <- piedata[piedata[,1] == i,]
    .tmp[piedata[,1] == i] <- tmp[,3]/sum(tmp[,3])
  }
  
  piedata<-cbind(.tmp, piedata)
  colnames(piedata) <- c("Summary", "Response", "Gender")
  piedata <- piedata[,-4]
  .df <- as.data.frame(piedata)
  library(grid)
  library(ggplot2)
  p = ggplot(data=.df, aes(x=factor(1), y=Summary, fill = factor(Gender)))     
  
  p = p + geom_bar(width = 1) + facet_grid(facets=. ~ Response) +  coord_polar(theta="y") 
  
  p = p + theme(axis.ticks=element_blank(), axis.title.x=element_blank(), axis.title.y=element_blank(), 
                axis.text.y=element_blank(), axis.text.x=element_text(face="bold", size=0), 
                plot.margin = unit(c(-.5,-.5,-.5,-.5), "lines"), panel.background = element_blank(), 
                strip.background=element_blank(), strip.text.x=element_blank())
  
  .color <- RColorBrewer::brewer.pal(8, 'Spectral')
  
  p = p + scale_fill_manual(values=.color)
  png(filename="Rplots.png",width=4000, height=400, res=300)
  print(p)
  dev.off(); R.utils::gzip.default("Rplots.png", overwrite=T, remove=T)
}



######################
## Loading Packages ##
######################
library(foreach)
library(polycor)
library(psych)
library(ltm)
library(scrime)
library(cluster)
library(stats)
library(mlogit)




##################################
## point biserial and extension ##  
##################################
## x : continuous variable ##
## y : nominal variable    ##
biserial.cor2 <- function (x, y, use = c("all.obs", "complete.obs"), level = 2)
{
  if (!is.numeric(x))
    stop("'x' must be a numeric variable.\n")
  y <- as.factor(y)
  stopifnot(length(levs <- levels(y)) >= 2)
  if (length(x) != length(y))
    stop("'x' and 'y' do not have the same length")
  use <- match.arg(use)
  if (use == "complete.obs") {
    cc.ind <- complete.cases(x, y)
    x <- x[cc.ind]
    y <- y[cc.ind]
  }
  ## Point Biserial ##
  if (length(levs) == 2) { 
    ind <- y == levs[level]
    diff.mu <- mean(x[ind]) - mean(x[!ind])
    prob <- mean(ind)
    diff.mu * sqrt(prob * (1 - prob))/sd(x)
    ## Point Biserial Extension ##
  } else if (length(levs) > 2) { 
    n.m <- table(y)
    n <- sum(n.m)
    x.bar <- mean(x)
    x.bar.m <- sapply(split(x, y), mean)
    s.sqr <-  sum((x - x.bar)^2/n) #(n-1)/n*var(y)
    s.m <- n.m*(x.bar.m-x.bar)/n
    n*sum(s.m^2/(n.m*s.sqr))
  }
}
pcc2 <- function (x, dist = FALSE, corrected = TRUE, version = 1) 
{
  minrc <- function(x){
    rx<-max(x,na.rm=TRUE)
    n.lev<-numeric(nrow(x))
    for(i in min(x,na.rm=TRUE):rx){
      tmp<-x==i
      tmp2<-rowSums(tmp,na.rm=TRUE)>0
      n.lev<-n.lev+tmp2
    }
    uni.lev<-sort(unique(n.lev),decreasing=TRUE)
    #if(uni.lev[1]!=rx)
    #        stop("No variable shows all possible levels.",call.=FALSE)
    if(length(uni.lev)==1)
      return(uni.lev)
    mat.rc<-matrix(rx,nrow(x),nrow(x))
    for(i in uni.lev[-1]){
      ids<-which(n.lev==i)
      mat.rc[ids,]<-i
      mat.rc[,ids]<-i
    }
    mat.rc
  }
  #if (!version %in% (1:3)) 
  #stop("version must be between 1 and 3.")
  if (dist & !corrected) 
    stop("corrected must be TRUE, if dist = TRUE.")
  chi2 <- scrime::rowChisqStats(x, compPval = FALSE)
  chi2 <- chi2 + t(chi2)
  if (any(is.na(x))) {
    naIdentifier <- !is.na(x)
    mat.n <- naIdentifier %*% t(naIdentifier)
  }
  else mat.n <- ncol(x)
  pcc <- chi2/(chi2 + mat.n)
  mat.rc <- minrc(x)
  mat.rc <- mat.rc/(mat.rc - 1)
  #print(mat.rc)
  #print(diag(pcc))
  diag(pcc) <- if (is.matrix(mat.rc)) 
    1/diag(mat.rc)
  else 1/mat.rc
  if (corrected) 
    pcc <- mat.rc * pcc
  colnames(pcc) <- rownames(pcc) <- rownames(x)
  if (!dist) {
    pcc <- sqrt(pcc)
    return(pcc)
  }
  pcc <- switch(version, sqrt(1 - pcc), 1 - sqrt(pcc), 1 - pcc)
  pcc
}

##############################################
## Function to establish Correlation Matrix ##                 
##############################################
calCor <- function(dat, types) {
  stopifnot(length(types)==2)
  
  if("con" %in% types) {
    if(all(types=="con")) {  ## con vs con: spearman's R
      dat <- na.omit(dat)
      res <- cor(as.numeric(dat[,1]), as.numeric(dat[,2]), use="pairwise.complete.obs",method="spearman")
    } else if("ord" %in% types) { ## con vs ord: polyserial
      dat <- na.omit(dat)
      res <- cor(as.numeric(dat[,1]), as.numeric(dat[,2]), use="pairwise.complete.obs",method="pearson")
    } else if("nom" %in% types) { ## con vs nom: point biserial and extension
      dat <- na.omit(dat)
      level <- nlevels(as.factor(dat[,which(types%in%c("nom"))]))-(max(dat[,which(types%in%c("nom"))])-min(dat[,which(types%in%c("nom"))])+1)
      value <- nlevels(as.factor(dat[,which(types%in%c("nom"))]))
      if((level != 0)|(value<2)|(length(which(dat[,which(types%in%c("nom"))]==1))<2)|(length(unique(dat[,which(types%in%c("con"))]))<3)){res <- NA}
      else{
        nom.idx <- which(types=="nom")
        res <- biserial.cor2(as.numeric(dat[,setdiff(1:ncol(dat),nom.idx)]),as.numeric(dat[,nom.idx]), use="complete.obs") #biserial or multivariate
      } }}
  else if("ord" %in% types) {
    if(all(types=="ord")) { ## ord vs ord: polychoric correlation and pearson's r 
      dat <- na.omit(dat)  
      res <- cor(as.numeric(dat[,1]), as.numeric(dat[,2]), use="pairwise.complete.obs",method="pearson")
    }
    else if("nom" %in% types) { #ord vs nom:rank biserial: point biserial and extension
      dat <- na.omit(dat)
      n1 <- as.factor(dat[,1])
      n2 <- as.factor(dat[,2])
      ord.idx <- which(types=="ord")
      nom.idx <- which(types=="nom")
      level.1 <- nlevels(as.factor(dat[,ord.idx]))-(max(dat[,ord.idx])-min(dat[,ord.idx])+1)
      value.1 <- nlevels(as.factor(dat[,ord.idx]))  
      level.2 <- nlevels(as.factor(dat[,nom.idx]))-(max(dat[,nom.idx])-min(dat[,nom.idx])+1)
      value.2 <- nlevels(as.factor(dat[,nom.idx]))
      if((level.2 != 0)|(value.1<3)|(level.1 != 0)|(value.2<2)|(length(which(dat[,which(types%in%c("nom"))]==1))<2)){res <- NA}
      else{
        n1 <- ord.idx
        ord <- as.factor(dat[,n1])
        dat[,n1] <- dat[,n1] - min(dat[,n1]) + 1
        dat.2 <- dat[,n1]
        n_old <- 0
        n <- vector(length=nlevels(ord))
        s <- min(dat[,n1])
        for (i in s:(nlevels(ord)+s-1)){
          no <- which(dat.2==i)
          n[i+1-s] <- length(no)
          new.level.1 <- 0.5*(1+n_old+n[i+1-s]+n_old)
          n_old <- sum(n[1:(i+1-s)])  
          dat[no,n1] <- new.level.1
        }
        dat[,1] <- factor(dat[,1]); dat[,2] <- factor(dat[,2])
        res <- biserial.cor2(as.numeric(dat[,setdiff(1:ncol(dat),ord.idx)]), dat[,ord.idx],use="complete.obs")} 
        #biserial or multivariat
    }
  }   
  else { ## nom vs nom: phi pearson's C and cramer's V
    stopifnot(all(types=="nom"))
    dat <- na.omit(dat)
    n1 <- as.factor(dat[,1]); n2 <- as.factor(dat[,2])
    level.1 <- nlevels(as.factor(dat[,1]))-(max(dat[,1])-min(dat[,1])+1)
    value.1 <- nlevels(as.factor(dat[,1]))  
    level.2 <- nlevels(as.factor(dat[,2]))-(max(dat[,2])-min(dat[,2])+1)
    value.2 <- nlevels(as.factor(dat[,2]))
    if((level.2 != 0)|(value.2<2)|(level.1 != 0)|(value.1<2)|(length(which(dat[,1]==1))<2)|(length(which(dat[,2]==1))<2)){res <- NA}
    else{
      if((nlevels(n1)==2 & nlevels(n2)==2)== TRUE) {
        dat <- as.data.frame(dat)
        res <- psych::phi(table(dat), digits=5)
      } 
      else if(nlevels(n1)!=nlevels(n2)){
        L <- min(nlevels(n1),nlevels(n2))
        dat[,1] <- dat[,1] - min(dat[,1]) + 1
        dat[,2] <- dat[,2] - min(dat[,2]) + 1
        chi.s <- scrime::rowChisqStats(t(as.matrix(dat)),compPval = FALSE)
        chi.s <- chi.s +t(chi.s)
        res <- (chi.s[2,1]/(nrow(dat)*ncol(dat)*(L-1)))^0.5}
      else 
      {dat <- as.data.frame(dat)
       res <- pcc2(t(dat))[2,1]
      }}
  }
  res
}


###################################
## Calculate the Distance matrix ##
###################################
mixedDist <- function(dat, types, isAbs=TRUE) {
  stopifnot(levels(factor(types))%in%c("con","ord","nom"))
  stopifnot(ncol(dat)==length(types))
  dist <- matrix(0,ncol(dat),ncol(dat))
  pairwise.cor = lapply(as.list(c(1:(ncol(dat)-1))),function(x) sapply(c((x+1):ncol(dat)),function(y) calCor(dat[,c(x,y)], types[c(x,y)])))
  for (i in 1:(ncol(dat)-1)) {
    dist[(i+1):ncol(dat),i] = (1 - pairwise.cor[[i]])/2
  }
  return(dist)
}
